/*
 * net/sched/sch_prio.c	Simple 3-band priority "scheduler".
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Authors:	Alexey Kuznetsov, <kuznet@ms2.inr.ac.ru>
 * Fixes:       19990609: J Hadi Salim <hadi@nortelnetworks.com>: 
 *              Init --  EINVAL when opt undefined
 */

#include <linux/config.h>
#include <linux/module.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/if_ether.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/notifier.h>
#include <net/ip.h>
#include <net/route.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <net/pkt_sched.h>

#include <linux/tcp.h>
#include <linux/udp.h>
#define DEBUG_ON 1
#if DEBUG_ON
#include <linux/proc_fs.h>

#define SCH_PRIO_PROC "qd_prio"

struct _sch_prio_cb{
        unsigned char trace_inQ;
        unsigned char trace_deQ;
};

static struct _sch_prio_cb g_prio_cb = {
        trace_inQ : 0,
        trace_deQ : 0,
};

static ssize_t 
proc_write_sch_prio_fops(struct file *filp,const char *buffer, size_t count , loff_t *offp)
{
        char line[10];
        copy_from_user(line,buffer,count);
        line[count]=0x00;

        if(count<1) 
                return count;
                
        switch(line[0]){
                case 'i':
                        {
                                if( line[1] == '1')
                                        g_prio_cb.trace_inQ = 1;
                                else
                                        g_prio_cb.trace_inQ = 0;
                        }
                        break;

                case 'd':
                        {
                                if( line[1] == '1')
                                        g_prio_cb.trace_deQ = 1;
                                else
                                        g_prio_cb.trace_deQ = 0;
                        }
                        break;
        }
        return count;
}
static ssize_t
proc_read_sch_prio_fops(struct file *filp, char *buf,size_t count , loff_t *offp)
{
        char tmp[128];
        int len=0;
        if(*offp!=0)
                return 0;
        sprintf(tmp,"trace_inQ=%d\ntrace_deQ=%d\n",
                        g_prio_cb.trace_inQ ,
                        g_prio_cb.trace_deQ
                        );

        len=strlen(tmp)+1;
        copy_to_user(buf,tmp,len);
        *offp = len;
        return len;
}

static struct proc_dir_entry *sch_prio_cfg;

static struct file_operations sch_prio_fops = {
read:   proc_read_sch_prio_fops,
write:  proc_write_sch_prio_fops,
};

#define TRACE(cond, format,argument...); if(cond) { printk(format,##argument); }
#else
#define TRACE(...);
#endif


struct prio_sched_data
{
	int bands;
	struct tcf_proto *filter_list;
	u8  prio2band[TC_PRIO_MAX+1];
	struct Qdisc *queues[TCQ_PRIO_BANDS];
#if DEBUG_ON
        int inq_err;
        int q_len[TCQ_PRIO_BANDS];
#endif

};


struct Qdisc *prio_classify(struct sk_buff *skb, struct Qdisc *sch,int *r)
{
	struct prio_sched_data *q = qdisc_priv(sch);
	u32 band = skb->priority;
	struct tcf_result res;
#if DEBUG_ON
        if( skb != NULL )
        {
                struct iphdr *ip = skb->nh.iph;
                struct tcphdr * tcp;
                struct udphdr * udp;

//              TRACE( g_prio_cb.trace_inQ, "\ninQ: pkt %p " , skb );
                TRACE( g_prio_cb.trace_inQ, "\ninQ: " );

                TRACE( g_prio_cb.trace_inQ, "L2 =%d " , ntohs(skb->protocol) );

                if(  ip != NULL )
                {

                        if( ip->protocol == IPPROTO_TCP  )
                        {
                                tcp = (struct tcphdr  *) ((char *)ip+(ip->ihl*4)) ;
                                TRACE( g_prio_cb.trace_inQ, "tcp port=%d -> %d" ,ntohs( tcp->source) , ntohs( tcp->dest));
                        }
                        else if (ip->protocol == IPPROTO_UDP )
                        {
                                udp = (struct udphdr  *) ((char *)ip+(ip->ihl*4)) ;
                                TRACE( g_prio_cb.trace_inQ, "udp port=%d -> %d" ,ntohs( udp->source) , ntohs( udp->dest) );
                        }
                        else
                        {
                                TRACE( g_prio_cb.trace_inQ, "L3 =%d " ,ntohs( skb->protocol ));
                        }


                        TRACE( g_prio_cb.trace_inQ, " IP Tos=0x%x Prio=0x%x Mark=0x%lx " , ntohs( ip->tos) , skb->priority, skb->nfmark );

                }
        }

#endif


	if (TC_H_MAJ(skb->priority) != sch->handle) {
#ifdef CONFIG_NET_CLS_ACT
		int result = 0, terminal = 0;
		result = tc_classify(skb, q->filter_list, &res);

		switch (result) {
			case TC_ACT_SHOT:
				*r = NET_XMIT_DROP;
				terminal = 1;
				break;
			case TC_ACT_STOLEN:
			case TC_ACT_QUEUED:
				terminal = 1;
				break;
			case TC_ACT_RECLASSIFY:
			case TC_ACT_OK:
			case TC_ACT_UNSPEC:
			default:
			break;
		};
		if (terminal) {
			kfree_skb(skb);
			TRACE( g_prio_cb.trace_inQ, "-> dropped\n");
			return NULL;
		} 

		if (!q->filter_list ) {
#else
		if (!q->filter_list || tc_classify(skb, q->filter_list, &res)) {
#endif
			if (TC_H_MAJ(band))
				band = 0;
#if DEBUG_ON
                        TRACE( g_prio_cb.trace_inQ, " -> Q[%d] (prio2band mapping) filter = %p\n", q->prio2band[band&TC_PRIO_MAX],q->filter_list  );

                        q->q_len[q->prio2band[band&TC_PRIO_MAX]]++;
#endif

			return q->queues[q->prio2band[band&TC_PRIO_MAX]];
		}
		band = res.classid;
	}
	band = TC_H_MIN(band) - 1;
	if (band > q->bands) 
	{
#if DEBUG_ON
                TRACE( g_prio_cb.trace_inQ, " -> Q[%d] (filter >  prio2band)\n", q->prio2band[0]);
                q->q_len[q->prio2band[0]]++;
#endif
		return q->queues[q->prio2band[0]];
	}
#if DEBUG_ON
        TRACE( g_prio_cb.trace_inQ, " -> Q[%d] (filter)\n", band);
        q->q_len[band]++;
#endif

	return q->queues[band];
}

static int
prio_enqueue(struct sk_buff *skb, struct Qdisc* sch)
{
	struct Qdisc *qdisc;
	int ret = NET_XMIT_SUCCESS;

	qdisc = prio_classify(skb, sch, &ret);

	if (NULL == qdisc)
		goto dropped;

	if ((ret = qdisc->enqueue(skb, qdisc)) == NET_XMIT_SUCCESS) {
		sch->stats.bytes += skb->len;
		sch->stats.packets++;
		sch->q.qlen++;
		return NET_XMIT_SUCCESS;
	}
        TRACE( g_prio_cb.trace_inQ, "inQ: fail !!\n");
#if DEBUG_ON
        {
                struct prio_sched_data *q = qdisc_priv(sch);
                q->inq_err ++;
        }
#endif

dropped:
#ifdef CONFIG_NET_CLS_ACT
	if (NET_XMIT_DROP == ret) {
#endif
		sch->stats.drops++;
#ifdef CONFIG_NET_CLS_ACT
	} else {
		sch->stats.overlimits++; /* abuse, but noone uses it */
	}
#endif
	return ret; 
}


static int
prio_requeue(struct sk_buff *skb, struct Qdisc* sch)
{
	struct Qdisc *qdisc;
	int ret = NET_XMIT_DROP;

	qdisc = prio_classify(skb, sch, &ret);
	if (qdisc == NULL)
		goto dropped;

	if ((ret = qdisc->ops->requeue(skb, qdisc)) == 0) {
		sch->q.qlen++;
		return 0;
	}
dropped:
	sch->stats.drops++;
	return NET_XMIT_DROP;
}


static struct sk_buff *
prio_dequeue(struct Qdisc* sch)
{
	struct sk_buff *skb;
	struct prio_sched_data *q = qdisc_priv(sch);
	int prio;
	struct Qdisc *qdisc;

	for (prio = 0; prio < q->bands; prio++) {
		qdisc = q->queues[prio];
		skb = qdisc->dequeue(qdisc);
		if (skb) {
#if DEBUG_ON
                        int idx;

                        TRACE( g_prio_cb.trace_deQ, "deQ: Q[%d] pkt %p [" , prio, skb );

                        for ( idx = 0; idx < q->bands; idx++)
                        {
                                TRACE( g_prio_cb.trace_deQ, "%d " , q->q_len[idx] );
                        }

                        TRACE( g_prio_cb.trace_deQ, "] *%d\n" , q->inq_err);
                        q->q_len[prio]--;
#endif
			sch->q.qlen--;
			return skb;
		}
	}
	return NULL;

}

static unsigned int prio_drop(struct Qdisc* sch)
{
	struct prio_sched_data *q = qdisc_priv(sch);
	int prio;
	unsigned int len;
	struct Qdisc *qdisc;

	for (prio = q->bands-1; prio >= 0; prio--) {
		qdisc = q->queues[prio];
		if ((len = qdisc->ops->drop(qdisc)) != 0) {
			sch->q.qlen--;
			return len;
		}
	}
	return 0;
}


static void
prio_reset(struct Qdisc* sch)
{
	int prio;
	struct prio_sched_data *q = qdisc_priv(sch);

	for (prio=0; prio<q->bands; prio++)
		qdisc_reset(q->queues[prio]);
	sch->q.qlen = 0;
}

static void
prio_destroy(struct Qdisc* sch)
{
	int prio;
	struct prio_sched_data *q = qdisc_priv(sch);
	struct tcf_proto *tp;

	while ((tp = q->filter_list) != NULL) {
		q->filter_list = tp->next;
		tcf_destroy(tp);
	}

	for (prio=0; prio<q->bands; prio++)
		qdisc_destroy(q->queues[prio]);
}

static int prio_tune(struct Qdisc *sch, struct rtattr *opt)
{
	struct prio_sched_data *q = qdisc_priv(sch);
	struct tc_prio_qopt *qopt = RTA_DATA(opt);
	int i;

	if (opt->rta_len < RTA_LENGTH(sizeof(*qopt)))
		return -EINVAL;
	if (qopt->bands > TCQ_PRIO_BANDS || qopt->bands < 2)
		return -EINVAL;

	for (i=0; i<=TC_PRIO_MAX; i++) {
		if (qopt->priomap[i] >= qopt->bands)
			return -EINVAL;
	}

	sch_tree_lock(sch);
	q->bands = qopt->bands;
	memcpy(q->prio2band, qopt->priomap, TC_PRIO_MAX+1);

	for (i=q->bands; i<TCQ_PRIO_BANDS; i++) {
		struct Qdisc *child = xchg(&q->queues[i], &noop_qdisc);
		if (child != &noop_qdisc)
			qdisc_destroy(child);
	}
	sch_tree_unlock(sch);

	for (i=0; i<=TC_PRIO_MAX; i++) {
		int band = q->prio2band[i];
		if (q->queues[band] == &noop_qdisc) {
			struct Qdisc *child;
			child = qdisc_create_dflt(sch->dev, &pfifo_qdisc_ops);
			if (child) {
				sch_tree_lock(sch);
				child = xchg(&q->queues[band], child);

				if (child != &noop_qdisc)
					qdisc_destroy(child);
				sch_tree_unlock(sch);
			}
		}
	}
	return 0;
}

static int prio_init(struct Qdisc *sch, struct rtattr *opt)
{
	struct prio_sched_data *q = qdisc_priv(sch);
	int i;

	for (i=0; i<TCQ_PRIO_BANDS; i++)
	{
		q->queues[i] = &noop_qdisc;
#if DEBUG_ON
                q->q_len[i]=0;
#endif
	}

#if DEBUG_ON
                q->inq_err=0;
#endif

	if (opt == NULL) {
		return -EINVAL;
	} else {
		int err;

		if ((err= prio_tune(sch, opt)) != 0)
			return err;
	}
	return 0;
}

static int prio_dump(struct Qdisc *sch, struct sk_buff *skb)
{
	struct prio_sched_data *q = qdisc_priv(sch);
	unsigned char	 *b = skb->tail;
	struct tc_prio_qopt opt;

	opt.bands = q->bands;
	memcpy(&opt.priomap, q->prio2band, TC_PRIO_MAX+1);
	RTA_PUT(skb, TCA_OPTIONS, sizeof(opt), &opt);
	return skb->len;

rtattr_failure:
	skb_trim(skb, b - skb->data);
	return -1;
}

static int prio_graft(struct Qdisc *sch, unsigned long arg, struct Qdisc *new,
		      struct Qdisc **old)
{
	struct prio_sched_data *q = qdisc_priv(sch);
	unsigned long band = arg - 1;

	if (band >= q->bands)
		return -EINVAL;

	if (new == NULL)
		new = &noop_qdisc;

	sch_tree_lock(sch);
	*old = q->queues[band];
	q->queues[band] = new;
	sch->q.qlen -= (*old)->q.qlen;
	qdisc_reset(*old);
	sch_tree_unlock(sch);

	return 0;
}

static struct Qdisc *
prio_leaf(struct Qdisc *sch, unsigned long arg)
{
	struct prio_sched_data *q = qdisc_priv(sch);
	unsigned long band = arg - 1;

	if (band >= q->bands)
		return NULL;

	return q->queues[band];
}

static unsigned long prio_get(struct Qdisc *sch, u32 classid)
{
	struct prio_sched_data *q = qdisc_priv(sch);
	unsigned long band = TC_H_MIN(classid);

	if (band - 1 >= q->bands)
		return 0;
	return band;
}

static unsigned long prio_bind(struct Qdisc *sch, unsigned long parent, u32 classid)
{
	return prio_get(sch, classid);
}


static void prio_put(struct Qdisc *q, unsigned long cl)
{
	return;
}

static int prio_change(struct Qdisc *sch, u32 handle, u32 parent, struct rtattr **tca, unsigned long *arg)
{
	unsigned long cl = *arg;
	struct prio_sched_data *q = qdisc_priv(sch);

	if (cl - 1 > q->bands)
		return -ENOENT;
	return 0;
}

static int prio_delete(struct Qdisc *sch, unsigned long cl)
{
	struct prio_sched_data *q = qdisc_priv(sch);
	if (cl - 1 > q->bands)
		return -ENOENT;
	return 0;
}


static int prio_dump_class(struct Qdisc *sch, unsigned long cl, struct sk_buff *skb,
			   struct tcmsg *tcm)
{
	struct prio_sched_data *q = qdisc_priv(sch);

	if (cl - 1 > q->bands)
		return -ENOENT;
	tcm->tcm_handle |= TC_H_MIN(cl);
	if (q->queues[cl-1])
		tcm->tcm_info = q->queues[cl-1]->handle;
	return 0;
}

static void prio_walk(struct Qdisc *sch, struct qdisc_walker *arg)
{
	struct prio_sched_data *q = qdisc_priv(sch);
	int prio;

	if (arg->stop)
		return;

	for (prio = 0; prio < q->bands; prio++) {
		if (arg->count < arg->skip) {
			arg->count++;
			continue;
		}
		if (arg->fn(sch, prio+1, arg) < 0) {
			arg->stop = 1;
			break;
		}
		arg->count++;
	}
}

static struct tcf_proto ** prio_find_tcf(struct Qdisc *sch, unsigned long cl)
{
	struct prio_sched_data *q = qdisc_priv(sch);

	if (cl)
		return NULL;
	return &q->filter_list;
}

static struct Qdisc_class_ops prio_class_ops = {
	.graft		=	prio_graft,
	.leaf		=	prio_leaf,
	.get		=	prio_get,
	.put		=	prio_put,
	.change		=	prio_change,
	.delete		=	prio_delete,
	.walk		=	prio_walk,
	.tcf_chain	=	prio_find_tcf,
	.bind_tcf	=	prio_bind,
	.unbind_tcf	=	prio_put,
	.dump		=	prio_dump_class,
};

static struct Qdisc_ops prio_qdisc_ops = {
	.next		=	NULL,
	.cl_ops		=	&prio_class_ops,
	.id		=	"prio",
	.priv_size	=	sizeof(struct prio_sched_data),
	.enqueue	=	prio_enqueue,
	.dequeue	=	prio_dequeue,
	.requeue	=	prio_requeue,
	.drop		=	prio_drop,
	.init		=	prio_init,
	.reset		=	prio_reset,
	.destroy	=	prio_destroy,
	.change		=	prio_tune,
	.dump		=	prio_dump,
	.owner		=	THIS_MODULE,
};

static int __init prio_module_init(void)
{
#if DEBUG_ON
        sch_prio_cfg=create_proc_entry(SCH_PRIO_PROC,0666,&proc_root);
        sch_prio_cfg->owner = THIS_MODULE;
        sch_prio_cfg->proc_fops = &sch_prio_fops;
#endif
	return register_qdisc(&prio_qdisc_ops);
}

static void __exit prio_module_exit(void) 
{
#if DEBUG_ON
        remove_proc_entry(SCH_PRIO_PROC,&proc_root);
#endif
	unregister_qdisc(&prio_qdisc_ops);
}

module_init(prio_module_init)
module_exit(prio_module_exit)

MODULE_LICENSE("GPL");
